#!/usr/bin/env python

# Reference: https://wiki.debian.org/LSBInitScripts

### BEGIN INIT INFO
# Provides:          tinker-access-client
# Required-Start:    $all
# Required-Stop:     $remote_fs $syslog
# Default-Start:     2 3 4 5
# Default-Stop:      0 1 6
# Short-Description: tinker-access-client
# Description:       The tinker-access-client service is responsible for coordinating communication
# between RPi's modules (i.e. RFID reader, LCD, power relays etc..) and the remote tinker_access_server.
### END INIT INFO

import os
import sys
from State import State
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientDaemon import ClientDaemon
from CommandHandler import CommandHandler
from ClientOptionParser import ClientOptionParser, Command


def __handle_status_command(opts, args):
    status = ClientDaemon.status(opts, args)
    if status:
        __handle_command_response(['{0}\n'.format(status)], 0)
    else:
        __handle_command_response(['{0}\n'.format(State.TERMINATED)], 1)


def __handle_update_command(opts, args):
    (messages, exit_code) = ClientDaemon.update(opts, args)
    __handle_command_response(messages, exit_code)


def __handle_start_command(opts, args):
    (messages, exit_code) = ClientDaemon.start(opts, args)
    __handle_command_response(messages, exit_code)


def __handle_command_response(messages, exit_code=0):
    if messages:
        for msg in messages:
            sys.stdout.write(msg)
            sys.stdout.flush()
    sys.exit(exit_code)


def run():
    if os.geteuid() != 0:
        sys.stdout.write(
            'You need to have root privileges to run {0} commands.\n'
            'Please try again, this time using \'sudo\'.\n'.format(PackageInfo.pip_package_name))
        sys.stdout.flush()
        sys.exit(1)

    (opts, args) = ClientOptionParser().parse_args()
    logger = ClientLogger.setup(opts)

    try:
        with CommandHandler(opts, args) as handler:
            handler.on(Command.STOP, ClientDaemon.stop)
            handler.on(Command.START, __handle_start_command)
            handler.on(Command.STATUS, __handle_status_command)
            handler.on(Command.UPDATE, __handle_update_command)
            handler.on(Command.REMOVE, ClientDaemon.remove)
            handler.on(Command.RESTART, ClientDaemon.restart)
            return handler.handle_command()

    except Exception as e:
        logger.exception(e)
        sys.stdout.write(str(e))
        sys.stdout.flush()
        sys.exit(1)

if __name__ == '__main__':
    run()



import os
import time

from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from CommandExecutor import CommandExecutor


class ServiceInstaller(object):
    def __init__(self, install_lib):
        self.__logger = ClientLogger.setup(phase='install')
        self.__command_executor = CommandExecutor(phase='install')

        self.__install_lib = install_lib
        self.__service_link = "/etc/init.d/{0}".format(PackageInfo.pip_package_name)
        self.__service_script = '{0}{1}/Service.py'.format(install_lib, PackageInfo.python_package_name)

    def install(self):
        try:
            self.__create_service()
            self.__configure_service()
            self.__restart_service()

        except Exception as e:
            self.__logger.debug('%s service installation failed.', PackageInfo.pip_package_name)
            self.__logger.exception(e)
            raise e

    def __create_service(self):
        self.__command_executor.ensure_execute_permission(self.__service_script)

        # remove any existing service if it is a file or directory, and it is not a symlink
        if os.path.exists(self.__service_link) and not os.path.islink(self.__service_link):
            os.remove(self.__service_link)

        # remove the existing service if it is a symlink and it is not pointed to the current target
        if os.path.lexists(self.__service_link) and os.readlink(self.__service_link) != self.__service_script:
            os.remove(self.__service_link)

        # create the symlink if it doesn't already exists
        if not os.path.lexists(self.__service_link):
            os.symlink(self.__service_script, self.__service_link)

    def __configure_service(self):
        time.sleep(5)
        self.__command_executor.execute_commands([
            'update-rc.d -f {0} defaults 91\n'.format(PackageInfo.pip_package_name)
        ])

    def __restart_service(self):
        time.sleep(5)
        self.__command_executor.execute_commands([
            'service {0} restart\n'.format(PackageInfo.pip_package_name)
        ])







# -*- coding: utf-8 -*-
# Original code found at:
# https://gist.github.com/DenisFromHR/cc863375a6e19dce359d

"""
Compiled, mashed and generally mutilated 2014-2015 by Denis Pleic
Made available under GNU GENERAL PUBLIC LICENSE

# Modified Python I2C library for Raspberry Pi
# as found on http://www.recantha.co.uk/blog/?p=4849
# Joined existing 'i2c_lib.py' and 'lcddriver.py' into a single library
# added bits and pieces from various sources
# By DenisFromHR (Denis Pleic)
# 2015-02-10, ver 0.1

"""

# Cobbled together from former TinkerAccess LcdApi.py and
# http://www.circuitbasics.com/raspberry-pi-i2c-lcd-set-up-and-programming/
# by kso512 (Chris Lindbergh) with lots of teaching from Erick McQueen
# 2017-05

import time
import smbus2 as smbus

# i2c bus (0 -- original Pi, 1 -- Rev 2 Pi)
I2CBUS = 1

class i2c_device:

    # Initialize I2C bus
    def __init__(self, addr, port=I2CBUS):
        self.addr = addr
        self.bus = smbus.SMBus(port)

    # Fix for 'too many files open' error
    def __del__(self):
        self.bus.close()

    # Fulfill context management requirements
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.bus.close()

    # Write a single command to the i2c bus
    def write_cmd(self, cmd):
        self.bus.write_byte_data(self.addr, cmd, 0)
        time.sleep(0.0001)

    # Write a command and argument to the i2c bus
    def write_cmd_arg(self, cmd, data):
        self.bus.write_byte_data(self.addr, cmd, data)
        time.sleep(0.0001)

    # Write a block of data to the i2c bus
    def write_block_data(self, cmd, data):
        self.bus.write_i2c_block_data(self.addr, cmd, data)
        time.sleep(0.0001)

    # Read a single byte from the i2c bus
    def read(self):
        return self.bus.read_byte(self.addr)

    # Read from the i2c bus
    def read_data(self, cmd):
        return self.bus.read_byte_data(self.addr, cmd)

    # Read a block of data from the i2c bus
    def read_block_data(self, cmd):
        return self.bus.read_block_data(self.addr, cmd)




import os
import time
import signal
import subprocess
from subprocess import CalledProcessError

from State import State
from Client import Client
from Command import Command
from daemonize import Daemonize
from PackageInfo import PackageInfo
from ClientOption import ClientOption
from ClientLogger import ClientLogger
from LoggedRequest import LoggedRequest
from CommandExecutor import CommandExecutor


# noinspection PyClassHasNoInit
class ClientDaemon:

    @staticmethod
    def start(opts, args):
        logger = ClientLogger.setup(opts)
        if not ClientDaemon.status(opts, args):
            pid_file = opts.get(ClientOption.PID_FILE)
            foreground = opts.get(ClientOption.DEBUG)

            def start():
                Client.run(opts, args)

            daemon = Daemonize(
                app=PackageInfo.pip_package_name,
                pid=pid_file,
                action=start,
                foreground=foreground,
                verbose=True,
                logger=logger,
                auto_close_fds=False
            )
            daemon.start()
            return [], 0
        else:
            return ['{0} is already running...\n'.format(PackageInfo.pip_package_name)], 1

    #TODO: need to cover scenario of machine is in use, and stop command is recieved

    @staticmethod
    def stop(opts, args):
        logger = ClientLogger.setup(opts)
        pid_file = opts.get(ClientOption.PID_FILE)
        logout_coast_time = opts.get(ClientOption.LOGOUT_COAST_TIME)

        try:
            for process_id in ClientDaemon.__get_process_ids():
                try:
                    os.kill(process_id, signal.SIGTERM)
                except Exception as e:
                    logger.exception(e)
        except CalledProcessError:
            pass

        # Wait for what we assume will be a graceful exit,
        # this will exit early if the pid file is removed as expected.
        current = time.time()
        max_wait_time = logout_coast_time + 5
        while time.time() - current < max_wait_time and os.path.isfile(pid_file):
            time.sleep(0.1)

        # This should never happen, but... if the pid file still exist at this point, we will nuke it from orbit!
        # Otherwise it will prevent the client daemon from starting again...
        if os.path.isfile(pid_file):
            try:
                os.remove(pid_file)
            except Exception as e:
                logger.exception(e)

    @staticmethod
    def update(opts, args):
        if not ClientDaemon.__is_in_use(opts, args):
            logger = ClientLogger.setup(opts)
            requested_version = args[1] if len(args) >= 2 else None
            if ClientDaemon.__should_update(opts, requested_version):
                try:
                    requested_package = PackageInfo.pip_package_name

                    # BUG: There is a big fat bug with pip that is causing it to redirect to
                    # install the latest version, even when a specific version is installed.
                    # I'll look into this when I get time.

                    if requested_version:
                        requested_package = '{0}=={1}'.format(requested_package, requested_version)

                    ClientDaemon.stop(opts, args)
                    CommandExecutor().execute_commands([
                        'pip install --upgrade --force-reinstall --ignore-installed --no-cache-dir {0}'
                        .format(requested_package)
                    ])
                except Exception as e:
                    msg = '{0} update failed, remediation maybe required!'.format(PackageInfo.pip_package_name)
                    logger.error(msg)
                    logger.exception(e)
                    raise e
                finally:
                    if not ClientDaemon.status(opts, args):
                        ClientDaemon.restart(opts, args)
            else:

                force_update_hint = 'Use the --force-update option to bypass ' \
                                    'the version check and re-install from PyPi.\n'\
                                    .format(
                                        PackageInfo.pip_package_name,
                                        PackageInfo.version
                                    )

                if not PackageInfo.version:
                    return [
                        'You are currently using a non-versioned build...\n',
                        force_update_hint
                    ], 1

                version = 'latest' if not requested_version else 'requested'
                return [
                    '{0} v{1} already matches the {2} version. \n'
                    'Use the --force-update option to bypass the version check and re-install.\n'.format(
                        PackageInfo.pip_package_name,
                        PackageInfo.version,
                        version
                    )], 1
        else:
            return ['{0} is currently in use, try again later...\n'.format(PackageInfo.pip_package_name)], 1

    @staticmethod
    def __get_latest_version_from_pypi():
        response = LoggedRequest.get('https://pypi.python.org/pypi/{0}/json'.format(PackageInfo.pip_package_name))
        response.raise_for_status()
        data = response.json()
        latest_version = data.get('info', {}).get('version', '0.0.0')
        return '.'.join([s.zfill(2) for s in latest_version.split('.')])

    @staticmethod
    def __should_update(opts, requested_version=None):
        logger = ClientLogger.setup(opts)
        try:
            if opts.get(ClientOption.FORCE_UPDATE):
                return True

            current_version = PackageInfo.version
            if current_version:
                if not requested_version:
                    return current_version != ClientDaemon.__get_latest_version_from_pypi()

                requested_version = '.'.join([s.zfill(2) for s in requested_version.split('.')])
            return current_version != requested_version
        except Exception as e:
            logger.exception(e)
        return False

    @staticmethod
    def restart(opts, args):
        logger = ClientLogger.setup(opts)
        reboot_delay = opts.get(ClientOption.REBOOT_DELAY)
        logout_coast_time = opts.get(ClientOption.LOGOUT_COAST_TIME)
        try:
            args[0] = Command.STOP.get('command')
            ClientDaemon.stop(opts, args)
            time.sleep(logout_coast_time)
            logger.debug('Restarting in %s seconds...', reboot_delay)
            time.sleep(reboot_delay)
            args[0] = Command.START.get('command')
            ClientDaemon.start(opts, args)
        except Exception as e:
            msg = '{0} restart failed.'.format(PackageInfo.pip_package_name)
            logger.error(msg)
            logger.exception(e)
            raise e

    @staticmethod
    def remove(opts, args):
        try:
            CommandExecutor().execute_commands([
                '{'
                '\tset +e',
                '\tservice tinker-access-client stop',
                '\tupdate-rc.d -f tinker-access-client remove',
                '\trm -rf /etc/init.d/tinker-access-client',
                '\twhile [[ $(pip uninstall tinker-access-client -y) == 0 ]]; do :; done',
                '\tfind /usr/local/lib/python2.7/dist-packages/ -name \'tinker[_-]access[_-]client*\' -type d -exec sudo rm -r "{}" \;',
                '\tsed -i.bak \'/^\/.*\/tinker_access_client$/s///g\' /usr/local/lib/python2.7/dist-packages/easy-install.pth',
                '} &> /dev/null'
            ])
        except Exception:
            pass

    @staticmethod
    def status(opts, _):
        status_file = opts.get(ClientOption.STATUS_FILE)

        process_ids = ClientDaemon.__get_process_ids()
        status = terminated = State.TERMINATED
        if os.path.isfile(status_file):
            with open(status_file, 'r') as f:
                status = f.readline().strip()

        return status if status is not terminated and len(process_ids) > 0 else None

    @staticmethod
    def __is_in_use(opts, args):
        status = ClientDaemon.status(opts, args)
        return status == State.IN_USE or status == State.IN_TRAINING

    @staticmethod
    def __get_process_ids():
        process_ids = []

        try:
            cmd = ['pgrep', '-f', '(/{0}\s+(start|restart|update))'.format(PackageInfo.pip_package_name)]
            for process_id in subprocess.check_output(cmd).splitlines():
                process_id = int(process_id)
                is_current_process = process_id == int(os.getpid())
                if not is_current_process:
                    process_ids.append(process_id)
        except CalledProcessError:
            pass

        return process_ids



import time
import math
import json
import serial
import logging
import threading
from LcdApi import LcdApi
from ClientOptionParser import ClientOption


class Channel(object):
    LCD, SERIAL, LED, PIN = range(0, 4)

    def __new__(cls, channel):
        for key, value in vars(Channel).items():
            if not key.startswith('__'):
                if value == channel:
                    return key
        return None


class DeviceApi(object):

    def __init__(self, opts):
        self.__opts = opts
        self.__fault = None
        self.__should_exit = False
        self.__edge_detected = False
        self.__first_line = ""
        self.__second_line = ""
        self.__lcd_refresh_timer = None
        self.__write_lcd_lock = threading.Lock()
        self.__logger = logging.getLogger(__name__)

    def __enter__(self):
        try:
            self.__configure_gpio()
            self.__configure_serial()
        except (RuntimeError, ImportError) as e:
            self.__logger.exception(e)
            raise e
        except Exception as e:
            self.__logger.debug(
                'Device initialization failed with %s.',
                json.dumps(self.__opts, indent=4, sort_keys=True))
            self.__logger.exception(e)
            raise e
        return self

    def __configure_gpio(self):
        # noinspection PyUnresolvedReferences
        import RPi.GPIO as GPIO
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.__opts.get(ClientOption.PIN_LED_RED), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_LED_BLUE), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_LED_GREEN), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_POWER_RELAY), GPIO.OUT)
        GPIO.setup(self.__opts.get(ClientOption.PIN_LOGOUT), GPIO.IN, GPIO.PUD_DOWN)

        # No pulldown for current sense pin, it will alter the RC time constant of the detect circuit
        GPIO.setup(self.__opts.get(ClientOption.PIN_CURRENT_SENSE), GPIO.IN, GPIO.PUD_OFF)

        if self.__opts.get(ClientOption.USE_ALARM):
            GPIO.setup(self.__opts.get(ClientOption.PIN_ALARM), GPIO.OUT)

        # If using programmable current threshold, set the pin as output high initially before PWM is enabled. If not using
        # the programmable threshold, set the pin to an input so it does not potentially contend with previous PCB revisions,
        # but enable the weak pullup so the threshold will float high, which will prevent the detection from triggering. 
        if self.__opts.get(ClientOption.USE_PGM_CURRENT_THRESHOLD):
            GPIO.setup(self.__opts.get(ClientOption.PIN_CURRENT_THRESHOLD), GPIO.OUT, initial=GPIO.HIGH)
        else:
            GPIO.setup(self.__opts.get(ClientOption.PIN_CURRENT_THRESHOLD), GPIO.IN, GPIO.PUD_UP)

        if self.__opts.get(ClientOption.USE_3V3_EN):
            # Toggle the 3.3V enable to reset I2C devices
            GPIO.setup(self.__opts.get(ClientOption.PIN_3V3_EN), GPIO.OUT)
            GPIO.output(self.__opts.get(ClientOption.PIN_3V3_EN), GPIO.LOW)
            time.sleep(1)
            GPIO.output(self.__opts.get(ClientOption.PIN_3V3_EN), GPIO.HIGH)
            time.sleep(2)

        if self.__opts.get(ClientOption.USE_ESTOP):
            if self.__opts.get(ClientOption.ESTOP_ACTIVE_HI):
                GPIO.setup(self.__opts.get(ClientOption.PIN_ESTOP), GPIO.IN, GPIO.PUD_UP)
            else:
                GPIO.setup(self.__opts.get(ClientOption.PIN_ESTOP), GPIO.IN, GPIO.PUD_DOWN)

        if self.__opts.get(ClientOption.USE_BYPASS_DETECT):
            GPIO.setup(self.__opts.get(ClientOption.PIN_BYPASS_DETECT), GPIO.IN, GPIO.PUD_UP)

        self.GPIO = GPIO

        self.__configure_pwm()

    def __configure_pwm(self):

        if self.__opts.get(ClientOption.USE_PGM_CURRENT_THRESHOLD):

            # Use pigpio for the current sense threshold PWM output as the RPi.GPIO PWM jitter is not very good
            import pigpio
            PIGPIO = pigpio.pi()

            # Set some parameters needed for PWM calcs
            pin_current_threshold = self.__opts.get(ClientOption.PIN_CURRENT_THRESHOLD)
            Rb = self.__opts.get(ClientOption.CT_BURDEN_RESISTOR)
            turns = self.__opts.get(ClientOption.CT_TURNS_RATIO)
            Vcc = 3.3

            # Attempt to set the PWM frequency and duty cycle range then get the actual values and adjust duty cycle to the real value
            PIGPIO.set_PWM_frequency(pin_current_threshold, 4000)
            PIGPIO.set_PWM_range(pin_current_threshold, 100)
            pwm_freq = PIGPIO.get_PWM_frequency(pin_current_threshold)
            pwm_range = PIGPIO.get_PWM_real_range(pin_current_threshold)
            PIGPIO.set_PWM_range(pin_current_threshold, pwm_range)

            max_threshold_in_amps_rms = Vcc / math.sqrt(2) / Rb * turns
            amps_resolution = max_threshold_in_amps_rms / pwm_range

            self.__logger.info('Current sense PWM setup: Rb: %d ohms, CT: %d:1, PWM Freq: %d, PWM Steps: %d, Max threshold: %.2f Amps(RMS), Step size: %.3f Amps(RMS).',
                               Rb, turns, pwm_freq, pwm_range, max_threshold_in_amps_rms, amps_resolution)

            pwm_dutycycle_raw = min(int(round(self.__opts.get(ClientOption.CURRENT_DETECT_SETTING) / 1000.0 / max_threshold_in_amps_rms * pwm_range)), pwm_range)
            pwm_dutycycle_pct = 100.0 * pwm_dutycycle_raw / pwm_range
            pwm_amps = pwm_dutycycle_raw * amps_resolution
            pwm_volts = Vcc * pwm_dutycycle_raw / pwm_range

            self.__logger.info('Current sense PWM setting: PWM Duty Cycle: %d (%.1f%%), PWM Output: %.3f Volts, Current Threshold Setting: %.2f Amps(RMS)',
                               pwm_dutycycle_raw, pwm_dutycycle_pct, pwm_volts, pwm_amps)

            PIGPIO.set_PWM_dutycycle(pin_current_threshold, pwm_dutycycle_raw)

            self.PIGPIO = PIGPIO

        else:
            self.PIGPIO = None

    def __configure_serial(self):
        serial_port_name = self.__opts.get(ClientOption.SERIAL_PORT_NAME)
        serial_port_speed = self.__opts.get(ClientOption.SERIAL_PORT_SPEED)
        self.__serial_connection = serial.Serial(serial_port_name, serial_port_speed)
        self.__serial_connection.flushInput()
        self.__serial_connection.flushOutput()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__stop()
        self.__do_cleanup()

    def __stop(self):
        self.__should_exit = True

    # noinspection PyBroadException
    def __do_cleanup(self):
        try:
            self.__write_to_led(True, False, False)
            self.__write_to_lcd('System Offline!', 'Try later...')
            self.__cancel_lcd_refresh_timer()
        except Exception:
            pass
        finally:
            if self.PIGPIO:
                self.PIGPIO.stop()
            self.GPIO.cleanup()

    def __do_callback(self, call_back, *args, **kwargs):
        try:
            call_back(*args, **kwargs)
        except Exception as e:
            self.__fault = e
            self.__stop()
        finally:
            self.__edge_detected = True

    def __poll_for_serial_input(self, call_back):
        while not self.__should_exit:
            try:
                badge_code = self.read(Channel.SERIAL)
                if badge_code:
                    self.__do_callback(call_back, badge_code=badge_code)
            except Exception as e:
                self.__fault = e
                self.__stop()
            time.sleep(.5)

    def __read_from_serial(self):
        serial_connection = self.__serial_connection

        if serial_connection.inWaiting() > 1:
            value = serial_connection.readline().strip()[-12:]
            serial_connection.flushInput()
            serial_connection.flushOutput()
            return value

        return None

    def __read_from_pin(self, pin, expected_state):
        # noinspection PyPep8Naming
        GPIO = self.GPIO
        expected_state = GPIO.LOW if not expected_state else GPIO.HIGH
        return GPIO.input(pin) == expected_state

    def __write_to_led(self, red, green, blue):
        # noinspection PyPep8Naming
        GPIO = self.GPIO
        GPIO.output(self.__opts.get(ClientOption.PIN_LED_RED), red)
        GPIO.output(self.__opts.get(ClientOption.PIN_LED_GREEN), green)
        GPIO.output(self.__opts.get(ClientOption.PIN_LED_BLUE), blue)

        # Retry backlight write multiple times, then raise error if unsuccessful
        num_attempts = 5
        for attempt in range(num_attempts):
            try:
                self.__call_lcd_backlight(red, green, blue)
            except Exception as e:
                if attempt == (num_attempts-1):
                    # All attempts exhausted, report as fatal
                    raise e
            else:
                break

    def __call_lcd_backlight(self, red, green, blue):
        try:
            with LcdApi(self.__opts, False) as lcd:
                lcd.rgb_backlight(red, green, blue)
        except Exception as e:
            self.__logger.debug('LCD I2C write backlight failed with colors \'%d %d %d\'.', red, green, blue)
            self.__logger.exception(e)
            raise e

    def __write_to_lcd(self, first_line, second_line, refresh=False):
        # Get the lock to allow only one thread to call the LCD write at a time
        self.__write_lcd_lock.acquire()

        # If not a refresh write, update the stored lines
        if not refresh:
            self.__first_line = first_line
            self.__second_line = second_line

        # Retry write multiple times, then raise error if unsuccessful
        num_attempts = 5
        for attempt in range(num_attempts):
            try:
                self.__call_lcd_write(self.__first_line, self.__second_line)
            except Exception as e:
                if attempt == (num_attempts-1):
                    # All attempts exhausted, report as fatal and make sure to release the lock
                    self.__write_lcd_lock.release()
                    raise e
            else:
                break

        self.__write_lcd_lock.release()

    def __call_lcd_write(self, first_line, second_line):
        try:
            self.__cancel_lcd_refresh_timer()
            with LcdApi(self.__opts) as lcd:
                lcd.write(first_line, second_line)
        except Exception as e:
            self.__logger.debug('LCD I2C write message failed with message \'%s %s\'.', first_line, second_line)
            self.__logger.exception(e)
            raise e

        self.__start_lcd_refresh_timer()

    def __cancel_lcd_refresh_timer(self):
        if self.__lcd_refresh_timer:
            self.__lcd_refresh_timer.cancel()

        self.__lcd_refresh_timer = None

    def __lcd_refresh_timer_tick(self):
        try:
            if not self.__should_exit:
                self.__write_to_lcd("", "", True)
        except Exception as e:
            self.__fault = e
            self.__stop()

    def __start_lcd_refresh_timer(self, interval=30):
        self.__cancel_lcd_refresh_timer()
        self.__lcd_refresh_timer = threading.Timer(
            interval,
            self.__lcd_refresh_timer_tick
        )
        self.__lcd_refresh_timer.start()

    def __write_to_pin(self, pin, state):
        # noinspection PyPep8Naming
        GPIO = self.GPIO
        state = GPIO.LOW if not state else GPIO.HIGH
        GPIO.output(pin, state)

    def on(self, channel, **kwargs):
        pin = kwargs.get('pin')
        direction = kwargs.get('direction')
        call_back = kwargs.get('call_back')

        # noinspection PyPep8Naming
        GPIO = self.GPIO
        if channel is Channel.PIN and pin and direction and call_back:

            def rising_edge_detected(*args, **kwargs):
                pin = args[0]
                time.sleep(.1)
                if GPIO.input(pin):
                    self.__do_callback(call_back, args, kwargs)

            def falling_edge_detected(*args, **kwargs):
                pin = args[0]
                time.sleep(.1)
                if not GPIO.input(pin):
                    self.__do_callback(call_back, args, kwargs)

            def both_edge_detected(*args, **kwargs):
                time.sleep(.5)
                self.__do_callback(call_back, args, kwargs)

            if direction == self.GPIO.RISING:
                GPIO.add_event_detect(pin, direction, callback=rising_edge_detected, bouncetime=500)
            elif direction == self.GPIO.FALLING:
                GPIO.add_event_detect(pin, direction, callback=falling_edge_detected, bouncetime=500)
            elif direction == self.GPIO.BOTH:
                GPIO.add_event_detect(pin, direction, callback=both_edge_detected, bouncetime=500 )
            else:
                raise NotImplementedError

        elif channel is Channel.SERIAL and direction is self.GPIO.IN and call_back:
            poll_for_serial_input = threading.Thread(
                name='poll_for_serial_input',
                target=self.__poll_for_serial_input,
                args=(call_back,)
            )
            poll_for_serial_input.daemon = True
            poll_for_serial_input.start()

        else:
            raise NotImplementedError

    def read(self, channel, *args):
        if self.__should_exit:
            return

        channel_name = Channel(channel)
        try:

            if channel == Channel.SERIAL:
                value = self.__read_from_serial()

            elif channel == Channel.PIN:
                pin = args[0] if len(args) >= 1 else None
                expected_state = args[1] if len(args) >= 2 else True
                value = self.__read_from_pin(pin, expected_state)

            else:
                raise NotImplementedError

        except Exception as e:
            self.__logger.debug('Read from %s failed with args \'%s\'.', channel_name, args)
            self.__logger.exception(e)
            raise e

        return value

    def write(self, channel, *args):
        if self.__should_exit:
            return

        channel_name = Channel(channel)
        try:
            if channel == Channel.LED:
                red = len(args) >= 1 and args[0] is True
                green = len(args) >= 2 and args[1] is True
                blue = len(args) >= 3 and args[2] is True
                self.__write_to_led(red, green, blue)

            elif channel == Channel.LCD:
                first_line = args[0] if len(args) >= 1 else ''
                second_line = args[1] if len(args) >= 2 else ''
                self.__write_to_lcd(first_line, second_line)

            elif channel == Channel.PIN:
                pin = args[0] if len(args) >= 1 else None
                state = args[1] if len(args) >= 2 else None
                self.__write_to_pin(pin, state)

            else:
                raise NotImplementedError

        except Exception as e:
            self.__logger.debug('Write to \'%s\' failed with args \'%s\'.', channel_name, args)
            self.__logger.exception(e)
            raise e

    def wait(self):
        while not self.__should_exit and not self.__edge_detected:
            time.sleep(1)
        self.__edge_detected = False
        self.__raise_fault()

    def __raise_fault(self):
        if self.__fault:
            raise self.__fault




import os
import logging
import logging.config as config
import logging.handlers as handlers
from ContextFilter import ContextFilter
from ClientOptionParser import ClientOptionParser, ClientOption


class ClientLogger(object):

    @staticmethod
    def setup(opts=None, phase=None):
        #TODO: remove ClientOptionParser, pass in opts always
        opts = opts if opts else ClientOptionParser(phase=phase).parse_args()[0]

        # if a logging config file exists, we will use that...
        logging_config_file = opts.get(ClientOption.LOGGING_CONFIG_FILE)
        if os.path.exists(logging_config_file):
            config.fileConfig(logging_config_file)
            return logging.getLogger()

        # otherwise add default handlers etc...
        logger = logging.getLogger()
        logger.filters = []
        logger.handlers = []

        log_level = opts.get(ClientOption.LOG_LEVEL)
        if phase == 'install' or opts.get(ClientOption.DEBUG):
            log_level = logging.DEBUG
        logger.setLevel(log_level)

        context_info = ' %(app_id)s (%(version)s) %(device_id)s %(device_name)s %(user_id)s %(user_name)s %(badge_code)s'

        # dev/log
        sys_log_file = '/var/log/syslog'
        if os.path.exists(sys_log_file):
            sys_log_handler = handlers.SysLogHandler('/dev/log')
            formatter = logging.Formatter('%(message)s' + context_info)
            sys_log_handler.setFormatter(formatter)
            sys_log_handler.addFilter(ContextFilter(opts))
            logger.addHandler(sys_log_handler)

        # console
        logger.addHandler(logging.StreamHandler())

        # fileHandler
        try:
            log_file = opts.get(ClientOption.LOG_FILE)
            if log_file:
                if not os.path.exists(os.path.dirname(log_file)):
                    os.makedirs(os.path.dirname(log_file))
                file_handler = handlers.TimedRotatingFileHandler(log_file, when='D', interval=1, backupCount=7)
                formatter = logging.Formatter('%(asctime)s %(message)s' + context_info, datefmt='%FT%T')
                file_handler.setFormatter(formatter)
                file_handler.addFilter(ContextFilter(opts))
                logger.addHandler(file_handler)
        except Exception as e:
            logger.exception(e)

        return logger

    # @staticmethod
    # def setup(phase=None):
    #     (opts, args) = ClientOptionParser(phase=phase).parse_args()
    #
    #     # if a logging config file exists, we will use that...
    #     logging_config_file = opts.get(ClientOption.LOGGING_CONFIG_FILE)
    #     if os.path.exists(logging_config_file):
    #         config.fileConfig(logging_config_file)
    #         return logging.getLogger()
    #
    #     # otherwise add default handlers etc...
    #     logger = logging.getLogger()
    #     logger.filters = []
    #     logger.handlers = []
    #
    #     log_level = opts.get(ClientOption.LOG_LEVEL)
    #     if phase == 'install' or opts.get(ClientOption.DEBUG):
    #         log_level = logging.DEBUG
    #     logger.setLevel(log_level)
    #
    #     context_info = ' %(app_id)s (%(version)s) %(device_id)s %(device_name)s %(user_id)s %(user_name)s %(badge_code)s'
    #
    #     # dev/log
    #     sys_log_file = '/var/log/syslog'
    #     if os.path.exists(sys_log_file):
    #         sys_log_handler = handlers.SysLogHandler('/dev/log')
    #         formatter = logging.Formatter('%(message)s' + context_info)
    #         sys_log_handler.setFormatter(formatter)
    #         sys_log_handler.addFilter(ContextFilter(opts))
    #         logger.addHandler(sys_log_handler)
    #
    #     # console
    #     logger.addHandler(logging.StreamHandler())
    #
    #     # fileHandler
    #     try:
    #         log_file = opts.get(ClientOption.LOG_FILE)
    #         if log_file:
    #             if not os.path.exists(os.path.dirname(log_file)):
    #                 os.makedirs(os.path.dirname(log_file))
    #             file_handler = handlers.TimedRotatingFileHandler(log_file, when='D', interval=1, backupCount=7)
    #             formatter = logging.Formatter('%(asctime)s %(message)s' + context_info, datefmt='%FT%T')
    #             file_handler.setFormatter(formatter)
    #             file_handler.addFilter(ContextFilter(opts))
    #             logger.addHandler(file_handler)
    #     except Exception as e:
    #         logger.exception(e)
    #
    #     return logger




class State(object):
    INITIALIZED = 'initialized'
    ESTOP = 'estop'
    BYPASSED = 'bypassed'
    IDLE = 'idle'
    UNLOCKED = 'unlocked'
    IN_USE = 'in_use'
    IN_TRAINING = 'in_training'
    TERMINATED = 'terminated'



# noinspection PyClassHasNoInit
class ClientOption:
    DEBUG = 'debug'
    PID_FILE = 'pid_file'
    LOG_FILE = 'log_file'
    LOG_LEVEL = 'log_level'
    DEVICE_ID = 'device_id'
    PIN_ESTOP = 'pin_estop'
    USE_ESTOP = 'use_estop'
    PIN_ALARM = 'pin_alarm'
    USE_ALARM = 'use_alarm'
    IS_A_DOOR = 'is_a_door'
    PIN_LOGOUT = 'pin_logout'
    PIN_3V3_EN = 'pin_3v3_en'
    USE_3V3_EN = 'use_3v3_en'
    CONFIG_FILE = 'config_file'
    PIN_LED_RED = 'pin_led_red'
    AUTO_UPDATE = 'auto_update'
    STATUS_FILE = 'status_file'
    PIN_LED_BLUE = 'pin_led_blue'
    REBOOT_DELAY = 'reboot_delay'
    FORCE_UPDATE = 'force_update'
    PIN_LED_GREEN = 'pin_led_green'
    SERVER_ADDRESS = 'server_address'
    DISPLAY_SERLCD = 'display_serlcd'
    CT_TURNS_RATIO = 'ct_turns_ratio'
    ESTOP_ACTIVE_HI = 'estop_active_hi'
    REBOOT_ON_ERROR = 'reboot_on_error'
    PIN_POWER_RELAY = 'pin_power_relay'
    SERIAL_PORT_NAME = 'serial_port_name'
    DOOR_UNLOCK_TIME = 'door_unlock_time'
    SERIAL_PORT_SPEED = 'serial_port_speed'
    PIN_CURRENT_SENSE = 'pin_current_sense'
    PIN_BYPASS_DETECT = 'pin_bypass_detect'
    USE_BYPASS_DETECT = 'use_bypass_detect'
    LOGOUT_COAST_TIME = 'logout_coast_time'
    CT_BURDEN_RESISTOR = 'ct_burden_resistor'
    LOGGING_CONFIG_FILE = 'logging_config_file'
    ALLOW_USER_OVERRIDE = 'allow_user_override'
    AUTO_UPDATE_INTERVAL = 'auto_update_interval'
    PIN_CURRENT_THRESHOLD = 'pin_current_threshold'
    DISABLE_TRAINING_MODE = 'disable_training_mode'
    DOOR_CONTINUOUS_UNLOCK = 'door_continuous_unlock'
    CURRENT_DETECT_SETTING = 'current_detect_setting'
    MAX_POWER_DOWN_TIMEOUT = 'max_power_down_timeout'
    USE_PGM_CURRENT_THRESHOLD = 'use_pgm_current_threshold'



import time
import datetime
import threading
from transitions import Machine

from State import State
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from DeviceApi import DeviceApi, Channel
from CommandExecutor import CommandExecutor
from AutoUpdateTimer import AutoUpdateTimer
from TinkerAccessServerApi import TinkerAccessServerApi
from ClientOptionParser import ClientOptionParser, ClientOption
from UserRegistrationException import UserRegistrationException
from UnauthorizedAccessException import UnauthorizedAccessException

maximum_lcd_characters = 16
training_mode_delay_seconds = 2
logout_timer_interval_seconds = 1


class Trigger(object):
    IDLE = 'idle'
    ESTOP = 'estop'
    BYPASS = 'bypass'
    UNLOCK = 'unlock'
    LOGIN = 'login'
    LOGOUT = 'logout'
    TERMINATE = 'terminate'


# noinspection PyUnusedLocal
class Client(Machine):
    def __init__(self, device=None, opts=None):
        self.__opts = opts
        self.__device = device
        self.__user_info = None
        self.__relock_timer = None
        self.__logout_timer = None
        self.__logout_timer_lock = threading.Lock()        
        self.__logger = ClientLogger.setup(opts)
        self.__tinkerAccessServerApi = TinkerAccessServerApi(opts)

        states = []
        for key, value in vars(State).items():
            if not key.startswith('__'):
                states.append(value)

        transitions = [
            {
                'source': [State.INITIALIZED, State.ESTOP, State.BYPASSED],
                'trigger': Trigger.IDLE,
                'dest': State.IDLE
            },

            {
                'source': [State.INITIALIZED, State.BYPASSED, State.IDLE, State.IN_USE, State.IN_TRAINING],
                'trigger': Trigger.ESTOP,
                'dest': State.ESTOP
            },

            {
                'source': [State.INITIALIZED, State.ESTOP, State.IDLE, State.IN_TRAINING],
                'trigger': Trigger.BYPASS,
                'dest': State.BYPASSED
            },
            
            {
                'source': [State.IDLE, State.IN_USE],
                'trigger': Trigger.UNLOCK,
                'dest': State.UNLOCKED,
                'conditions': ['is_normal_hours']
            },

            {
                'source': [State.IDLE],
                'trigger': Trigger.LOGIN,
                'dest': State.IN_USE,
                'conditions': ['is_authorized']
            },

            {
                'source': [State.IN_USE],
                'trigger': Trigger.LOGIN,
                'dest': State.IN_USE,
                'conditions': ['should_extend_current_session']
            },

            {
                'source': [State.UNLOCKED, State.IN_USE, State.IN_TRAINING],
                'trigger': Trigger.LOGOUT,
                'dest': State.IDLE
            },

            {
                'source': [State.IDLE, State.ESTOP, State.BYPASSED],
                'trigger': Trigger.LOGOUT,
                'dest': State.IN_TRAINING,
                'conditions': ['is_waiting_for_training']
            },

            {
                'source': '*',
                'trigger': Trigger.TERMINATE,
                'dest': State.TERMINATED
            }
        ]

        Machine.__init__(self, queued=True, states=states, ignore_invalid_triggers=True,
                         transitions=transitions, initial=State.INITIALIZED, after_state_change='update_status')

    def __enter__(self):
        self.update_status()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()

    #
    # estop -- The machine is in e-stop and waiting e-stop to be cleared or enter training mode
    #

    def __ensure_estop(self):
        self.__do_logout()
        self.__disable_power()
        self.__show_red_led()
        self.__show_estop_activated()

    def __show_estop_activated(self):
        self.__device.write(
            Channel.LCD,
            'E-STOP ACTIVATED'.center(maximum_lcd_characters, ' '),
            'RESET THE SWITCH'.center(maximum_lcd_characters, ' ')
        )

    #
    # bypassed -- The machine has been bypassed and waiting for bypass to be cleared or enter training mode
    #

    def __ensure_bypass(self):
        self.__do_logout()
        self.__disable_power()
        self.__show_yellow_led()
        self.__show_bypassed()

    def __show_yellow_led(self):
        self.__device.write(Channel.LED, True, True, False)
        self.__set_alarm_output(False)

    def __show_bypassed(self):
        self.__device.write(
            Channel.LCD,
            'TINKERACCESS'.center(maximum_lcd_characters, ' '),
            'IS BYPASSED'.center(maximum_lcd_characters, ' ')
        )

    #
    # idle -- The machine is idle and waiting for a badge to be scanned
    #

    def __ensure_idle(self):
        self.__do_logout()
        self.__disable_power()
        self.__show_blue_led()
        self.__show_scan_badge()

    def __do_login(self, override, *args, **kwargs):
        badge_code = kwargs.get('badge_code')

        # noinspection PyBroadException
        try:
            self.__show_attempting_login(1)
            self.__update_user_context(
                self.__tinkerAccessServerApi.login(badge_code)
            )
            remaining_seconds = self.__user_info.get('remaining_seconds')
            self.__show_access_granted(1)

        except UnauthorizedAccessException as e:
            self.__handle_unauthorized_access_exception()
            if not override:
                self.__ensure_idle()
            else:
                self.__ensure_in_use()

        except Exception as e:
            self.__handle_unexpected_exception()
            if not override:
                self.__ensure_idle()
            else:
                self.__ensure_in_use()

        return self.__user_info is not None

    def __show_attempting_login(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ATTEMPTING'.center(maximum_lcd_characters, ' '),
            'LOGIN...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __update_user_context(self, user_info):
        self.__user_info = user_info
        for handler in self.__logger.handlers:
            for context_filter in handler.filters:
                update_user_context = getattr(context_filter, "update_user_context", None)
                if callable(update_user_context):
                    context_filter.update_user_context(self.__user_info)

    def __handle_unauthorized_access_exception(self):
        self.__show_red_led()
        self.__show_access_denied(2)

    def __show_access_denied(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ACCESS DENIED'.center(maximum_lcd_characters, ' '),
            'TAKE THE CLASS'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_red_led(self):
        self.__device.write(Channel.LED, True, False, False)
        self.__set_alarm_output(True)

    def __set_alarm_output(self, state):
        if self.__opts.get(ClientOption.USE_ALARM):
            self.__device.write(Channel.PIN, self.__opts.get(ClientOption.PIN_ALARM), state)

    def __do_logout(self):
        self.__cancel_logout_timer()
        self.__cancel_relock_timer()

        if self.__user_info:
            badge_code = self.__user_info.get('badge_code')
            threading.Timer(0, self.__tinkerAccessServerApi.logout, (badge_code, )).start()

        self.__update_user_context(None)

    def __show_blue_led(self):
        self.__device.write(Channel.LED, False, False, True)
        self.__set_alarm_output(False)

    def __show_scan_badge(self):
        self.__device.write(
            Channel.LCD,
            'SCAN BADGE'.center(maximum_lcd_characters, ' '),
            'TO LOGIN'.center(maximum_lcd_characters, ' ')
        )

    #
    # unlocked -- The door is held manually in a continuous unlocked state
    #

    def __ensure_unlocked(self):
        self.__do_logout()
        self.__enable_power()
        self.__show_green_led()
        self.__show_unlocked()

    def __show_unlocked(self):
        self.__device.write(
            Channel.LCD,
            'TINKERACCESS'.center(maximum_lcd_characters, ' '),
            'IS UNLOCKED'.center(maximum_lcd_characters, ' ')
        )
        
    def __start_relock_timer(self):
        self.__cancel_relock_timer()
        self.__relock_timer = threading.Timer(
            60,
            self.__relock_timer_tick
        )
        self.__relock_timer.start()

    def __relock_timer_tick(self):
        try:
            if not self.is_terminated() and self.__relock_timer:
                if not self.is_normal_hours():
                    # Outside of normal hours, relock the door
                    self.logout()
                    return

                self.__start_relock_timer()
        except Exception as e:
            raise e

    def __cancel_relock_timer(self):
        try:
            if self.__relock_timer:
                self.__relock_timer.cancel()
        except Exception as e:
            raise e
        finally:
            self.__relock_timer = None

    #
    # in_use -- The machine is currently in use and the logout timer is ticking...
    #

    def __ensure_in_use(self):
        self.__enable_power()
        self.__show_green_led()

    def __enable_power(self):
        self.__device.write(Channel.PIN, self.__opts.get(ClientOption.PIN_POWER_RELAY), True)

    def __show_access_granted(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ACCESS GRANTED'.center(maximum_lcd_characters, ' '),
        )
        time.sleep(delay)

    def __show_green_led(self):
        self.__device.write(Channel.LED, False, True, False)
        self.__set_alarm_output(False)

    def __start_logout_timer(self, refresh=False):
        self.__cancel_logout_timer(refresh)
        self.__logout_timer = threading.Timer(
            logout_timer_interval_seconds,
            self.__logout_timer_tick
        )
        self.__logout_timer.start()

    def __logout_timer_tick(self):
        self.__logout_timer_lock.acquire()
        thread_locked = True

        try:
            if not self.is_terminated() and self.__user_info and self.__logout_timer:
                remaining_seconds = self.__user_info.get('remaining_seconds')

                if remaining_seconds <= 0:
                    # Release the lock before triggering logout
                    self.__logout_timer_lock.release()
                    thread_locked = False
                    self.logout()
                    return

                self.__user_info['remaining_seconds'] = (remaining_seconds - logout_timer_interval_seconds)
                self.__show_remaining_time()
                self.__start_logout_timer(True)
        except Exception as e:
            raise e
        finally:
            if thread_locked:
                self.__logout_timer_lock.release()

    def __show_remaining_time(self):
        remaining_seconds = self.__user_info.get('remaining_seconds')
        if remaining_seconds < 300:
            self.__toggle_red_led()

        m, s = divmod(int(remaining_seconds), 60)
        h, m = divmod(m, 60)
        user_name = self.__user_info.get('user_name')
        self.__device.write(
            Channel.LCD,
            user_name.center(maximum_lcd_characters, ' '),
            '{0:02d}:{1:02d}:{2:02d}'.format(h, m, s).center(maximum_lcd_characters, ' ')
        )

    def __toggle_red_led(self):
        red_led_status = self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_LED_RED))
        self.__device.write(Channel.LED, not red_led_status, False, False)
        self.__set_alarm_output(True)

    def __cancel_logout_timer(self, refresh=False):
        if not refresh:
            self.__logout_timer_lock.acquire()

        try:
            if self.__logout_timer:
                self.__logout_timer.cancel()
        except Exception as e:
            raise e
        finally:
            self.__logout_timer = None
            if not refresh:
                self.__logout_timer_lock.release()

    def __extend_session(self):
        self.__cancel_logout_timer()

        # TODO: add api call to let server know that time has been extended...

        session_seconds = self.__user_info.get('session_seconds')
        remaining_seconds = self.__user_info.get('remaining_seconds')
        remaining_extensions = self.__user_info.get('remaining_extensions')

        if remaining_extensions:
            if remaining_extensions != float('inf'):
                self.__user_info['remaining_extensions'] = remaining_extensions - 1

            remaining_seconds = remaining_seconds + session_seconds
            self.__user_info['remaining_seconds'] = remaining_seconds
            self.__logger.info('Session extended %s remaining_seconds', remaining_seconds)

            self.__show_session_extended(1)
        else:
            self.__show_no_extensions_remaining(2)

        self.__start_logout_timer()

    def __show_session_extended(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'SESSION EXTENDED'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
        self.__show_remaining_time()

    def __show_no_extensions_remaining(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'NO EXTENSIONS'.center(maximum_lcd_characters, ' '),
            'REMAINING...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
        self.__show_remaining_time()

    #
    # training - the client has entered training mode
    #

    def __ensure_training_mode(self):
        self.__do_logout()
        self.__disable_power()
        self.__show_magenta_led()
        self.__show_training_mode_activated(1)

    def __show_magenta_led(self):
        self.__device.write(Channel.LED, True, False, True)
        self.__set_alarm_output(False)

    def __show_training_mode_activated(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'TRAINING MODE'.center(maximum_lcd_characters, ' '),
            'ACTIVATED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __activate_trainer(self, badge_code):
        # Note: currently we call the normal login method on the tinkerAccessServer
        # the backend ensures the user is a trainer on the 'registerUser' call.
        # I expect this might change in the future once the server code gets refactored
        try:
            self.__show_attempting_login(1)
            self.__update_user_context(
                self.__tinkerAccessServerApi.login(badge_code)
            )
            self.__show_trainer_accepted(1)
            self.__prompt_for_student_badge()

        except UnauthorizedAccessException:
            self.__handle_unauthorized_access_exception()
            self.__prompt_for_trainer_badge()
        except Exception as e:
            self.__handle_unexpected_exception()
            self.__prompt_for_trainer_badge()

        return self.__user_info is not None

    def __prompt_for_trainer_badge(self):
        self.__show_magenta_led()
        self.__show_scan_trainer_badge()

    def __show_scan_trainer_badge(self):
        self.__device.write(
            Channel.LCD,
            'SCAN'.center(maximum_lcd_characters, ' '),
            'TRAINER BADGE...'.center(maximum_lcd_characters, ' ')
        )

    def __show_trainer_accepted(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'TRAINER'.center(maximum_lcd_characters, ' '),
            'ACCEPTED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __prompt_for_student_badge(self):
        self.__show_magenta_led()
        self.__show_scan_student_badge()

    def __show_scan_student_badge(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'SCAN'.center(maximum_lcd_characters, ' '),
            'STUDENT BADGE...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __register_student(self, badge_code):
        try:
            self.__show_attempting_registration(1)
            trainer_id = self.__user_info.get('user_id')
            trainer_badge_code = self.__user_info.get('badge_code')
            self.__tinkerAccessServerApi.register_user(trainer_id, trainer_badge_code, badge_code)
            self.__show_student_registered(1)
        except UserRegistrationException:
            self.__handle_user_registration_exception()
            self.__show_invalid_user(2)
        except Exception as e:
            self.__handle_user_registration_exception()
            self.__handle_unexpected_exception()
        finally:
            self.__prompt_for_student_badge()

    def __show_attempting_registration(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'ATTEMPTING'.center(maximum_lcd_characters, ' '),
            'REGISTRATION...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_student_registered(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'STUDENT'.center(maximum_lcd_characters, ' '),
            'REGISTERED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __handle_user_registration_exception(self):
        self.__show_red_led()
        self.__show_registration_failed(2)

    def __show_registration_failed(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'REGISTRATION'.center(maximum_lcd_characters, ' '),
            'FAILED...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_invalid_user(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'INVALID'.center(maximum_lcd_characters, ' '),
            'USER...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
        
    #
    # logout/terminated - the user has logged out, or the client is shutting down
    #

    def __disable_power(self):
        power_relay_pin = self.__opts.get(ClientOption.PIN_POWER_RELAY)
        if self.__device.read(Channel.PIN, power_relay_pin):
            is_machine_running = self.__wait_for_power_down()
            if is_machine_running:
                self.__wait_for_logout_coast_time()

        if self.__device.read(Channel.PIN, power_relay_pin):
            self.__device.write(Channel.PIN, power_relay_pin, False)
            self.__show_disabling_power()

    def __wait_for_power_down(self):
        power_relay_pin = self.__opts.get(ClientOption.PIN_POWER_RELAY)
        current_sense_pin = self.__opts.get(ClientOption.PIN_CURRENT_SENSE)
        max_power_down_timeout = self.__opts.get(ClientOption.MAX_POWER_DOWN_TIMEOUT)

        if max_power_down_timeout is None:
            max_power_down_timeout = float('inf')

        current = time.time()
        is_machine_running = False
        while time.time() - current < max_power_down_timeout and self.__device.read(Channel.PIN, current_sense_pin):
            is_machine_running = True
            self.__show_red_led()
            self.__show_waiting_for_power_down()
            time.sleep(0.5)

        return is_machine_running

    def __wait_for_logout_coast_time(self):
        logout_coast_time = self.__opts.get(ClientOption.LOGOUT_COAST_TIME)
        if logout_coast_time:
            self.__show_coasting_down()
            time.sleep(logout_coast_time)

    def __show_disabling_power(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'DISABLING'.center(maximum_lcd_characters, ' '),
            'POWER...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_waiting_for_power_down(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'WAITING FOR ...'.center(maximum_lcd_characters, ' '),
            'MACHINE TO STOP'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_coasting_down(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'COASTING'.center(maximum_lcd_characters, ' '),
            'DOWN...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __handle_unexpected_exception(self):
        self.__show_red_led()
        self.__show_error_occurred(2)
        self.__show_please_try_again(2)

    def __show_error_occurred(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'THERE WAS AN'.center(maximum_lcd_characters, ' '),
            'UNEXPECTED ERROR'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)

    def __show_please_try_again(self, delay=0):
        self.__device.write(
            Channel.LCD,
            'PLEASE'.center(maximum_lcd_characters, ' '),
            'TRY AGAIN...'.center(maximum_lcd_characters, ' ')
        )
        time.sleep(delay)
        
    #
    # a badge code has been detected on the serial input
    #

    def handle_badge_code(self, *args, **kwargs):
        badge_code = kwargs.get('badge_code')
        if self.state == State.IN_TRAINING:
            if not self.__user_info:
                if self.__activate_trainer(badge_code):
                    self.__show_scan_student_badge()
                else:
                    self.__show_scan_trainer_badge()
            else:
                if not self.__is_current_badge_code(*args, **kwargs):
                    self.__register_student(badge_code)
        else:
            self.login(*args, **kwargs)

    #
    # logout_detected - logout button detected
    #

    def logout_detected(self, *args, **kwargs):
        if self.is_estop_activated() and (self.state == State.IN_TRAINING):
            # Call estop() if exiting training and estop button is active to trigger return to estop state
            self.estop()
        elif self.is_bypass_detected() and (self.state == State.IN_TRAINING):
            # Call bypass() if exiting training and bypass is detected to trigger return to bypassed state
            self.bypass()
        elif (self.__opts.get(ClientOption.IS_A_DOOR) and self.__opts.get(ClientOption.DOOR_CONTINUOUS_UNLOCK) and
              ((self.state == State.IDLE) or (self.state == State.IN_USE))):
            # Call unlock() if continuous unlock is enabled and in idle or in_use state
            self.unlock()
        else:
            # Otherwise call logout() to return to idle
            self.logout()

    #
    # estop_change - change in status of e-stop input
    #

    def estop_change(self, *args, **kwargs):
        if self.is_estop_activated():
            # E-Stop pushbutton was pressed
            if self.state != State.IN_TRAINING:
                # Do not call estop() from in_training state, wait to call it upon exit from state
                self.estop()
        elif self.state == State.ESTOP:
            # E-Stop pushbutton was reset and in ESTOP state, wait for a potential bypass detect
            time.sleep(0.5)
            if self.is_bypass_detected():
                # Bypass detected so transition to bypass mode instead
                self.bypass()
            else:
                # E-Stop was reset and no bypass detected so transition to idle
                self.idle()

    #
    # bypass_change - change in status of bypass detect input
    #

    def bypass_change(self, *args, **kwargs):
        if self.is_bypass_detected():
            # Bypass was detected
            if self.state == State.IDLE:
                # Do not call bypass() from in_training state, wait to call it upon exit from state
                # Only trigger positive edge change of bypass detect from IDLE state
                self.bypass()
        elif self.state == State.BYPASSED:
            # Bypass detect was cleared and in BYPASSED state, trigger return to IDLE
            self.idle()

    #
    # wait - wait for the next edge detection event from the device.
    #

    def wait(self):
        self.__device.wait()

    #
    # status - The client has received a status command
    #

    def status(self):
        return self.state

    def update_status(self, *args, **kwargs):
        status_file = self.__opts.get(ClientOption.STATUS_FILE)
        with open(status_file, 'w') as f:
            f.write('{0}\n'.format(self.status()))
            f.close()
    #
    # conditions - used to allow/prevent triggers causing a transition if the conditions are not met.
    #

    def is_estop_activated(self):
        return self.__opts.get(ClientOption.USE_ESTOP) and (
            (self.__opts.get(ClientOption.ESTOP_ACTIVE_HI) and self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_ESTOP))) or
            (not self.__opts.get(ClientOption.ESTOP_ACTIVE_HI) and not self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_ESTOP)))
        )

    def is_bypass_detected(self):
        return self.__opts.get(ClientOption.USE_BYPASS_DETECT) and self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_BYPASS_DETECT))

    def is_in_use(self):
        return self.status() == State.IN_USE or self.state == State.IN_TRAINING

    def is_normal_hours(self):
        now = datetime.datetime.now().time()
        start = datetime.time(7, 30)
        end = datetime.time(22, 00)
        return (start <= now <= end)
    
    def is_terminated(self):
        return self.status() == State.TERMINATED

    def is_authorized(self, *args, **kwargs):
        return self.__do_login(False, *args, **kwargs)

    def is_waiting_for_training(self, *args, **kwargs):
        # Do not enter training mode if it is disabled in the config file, or a continuous unlock mode door is defined
        if ((self.__opts.get(ClientOption.IS_A_DOOR) and self.__opts.get(ClientOption.DOOR_CONTINUOUS_UNLOCK)) or
            self.__opts.get(ClientOption.DISABLE_TRAINING_MODE)):
            return False
        else:
            current = time.time()
            while (not self.is_terminated() and time.time() - current < training_mode_delay_seconds and
                   self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_LOGOUT))):
                time.sleep(0.1)

            return self.__device.read(Channel.PIN, self.__opts.get(ClientOption.PIN_LOGOUT))

    def should_extend_current_session(self, *args, **kwargs):
        if self.__is_current_badge_code(*args, **kwargs):
            # Same user badge, extend session
            self.__extend_session()
            return True
        elif self.__opts.get(ClientOption.ALLOW_USER_OVERRIDE):
            # Different user badge and override allowed, attempt new login
            self.__cancel_logout_timer()
            self.__do_login(True, *args, **kwargs)
            return True
        else:
            return False

    def __is_current_badge_code(self, *args, **kwargs):
        new_badge_code = kwargs.get('badge_code')
        current_badge_code = self.__user_info.get('badge_code') if self.__user_info else None

        if current_badge_code and current_badge_code == new_badge_code:
            return True

        return False

    def on_enter_estop(self, *args, **kwargs):
        self.__ensure_estop()
        self.__logger.warning('Emergency Stop Detected')

    def on_enter_bypassed(self, *args, **kwargs):
        self.__ensure_bypass()
        self.__logger.warning('TinkerAccess has been Bypassed')

    def on_enter_idle(self, *args, **kwargs):
        self.__ensure_idle()
        # Wait to make sure bypass is not detected
        time.sleep(0.5)
        if self.is_bypass_detected():
            self.bypass()

    def on_enter_unlocked(self, *args, **kwargs):
        self.__ensure_unlocked()
        self.__start_relock_timer()
        self.__logger.info('TinkerAccess has been Unlocked')

    def on_enter_in_use(self, *args, **kwargs):
        self.__ensure_in_use()
        self.__start_logout_timer()

    def on_enter_in_training(self, *args, **kwargs):
        self.__ensure_training_mode()
        self.__show_scan_trainer_badge()

    def on_enter_terminated(self, *args, **kwargs):
        self.__ensure_idle()

    @staticmethod
    def run(opts, args):
        logger = ClientLogger.setup(opts)
        reboot_delay = opts.get(ClientOption.REBOOT_DELAY) * 60
        reboot_on_error = opts.get(ClientOption.REBOOT_ON_ERROR)

        try:
            with DeviceApi(opts) as device, \
                    Client(device, opts) as client, \
                    AutoUpdateTimer(client, opts) as auto_update_timer:

                device.on(
                    Channel.SERIAL,
                    direction=device.GPIO.IN,
                    call_back=client.handle_badge_code
                )

                device.on(
                    Channel.PIN,
                    pin=opts.get(ClientOption.PIN_LOGOUT),
                    direction=device.GPIO.RISING,
                    call_back=client.logout_detected
                )

                if opts.get(ClientOption.USE_ESTOP):

                    device.on(
                        Channel.PIN,
                        pin=opts.get(ClientOption.PIN_ESTOP),
                        direction=device.GPIO.BOTH,
                        call_back=client.estop_change
                    )

                if opts.get(ClientOption.USE_BYPASS_DETECT):

                    device.on(
                        Channel.PIN,
                        pin=opts.get(ClientOption.PIN_BYPASS_DETECT),
                        direction=device.GPIO.BOTH,
                        call_back=client.bypass_change
                    )

                if client.is_estop_activated():
                    client.estop()
                elif client.is_bypass_detected():
                    client.bypass()
                else:
                    client.idle()

                auto_update_timer.start()
                while not client.is_terminated():
                    logger.debug('%s is waiting...', PackageInfo.pip_package_name)
                    client.wait()

        except (KeyboardInterrupt, SystemExit) as e:
            pass

        except Exception as e:
            logger.exception(e)

            if reboot_on_error:

                # reboot is only supported on Raspberry PI devices
                # noinspection PyBroadException
                try:
                    # noinspection PyUnresolvedReferences
                    import RPi.GPIO
                    logger.error('Rebooting in %s minutes...', reboot_delay / 60)
                    CommandExecutor().execute_commands([
                        'sleep {0}s'.format(reboot_delay),
                        'reboot now'
                    ])
                except Exception:
                    pass




# -*- coding: utf-8 -*-
# Original code found at:
# https://gist.github.com/DenisFromHR/cc863375a6e19dce359d

"""
Compiled, mashed and generally mutilated 2014-2015 by Denis Pleic
Made available under GNU GENERAL PUBLIC LICENSE

# Modified Python I2C library for Raspberry Pi
# as found on http://www.recantha.co.uk/blog/?p=4849
# Joined existing 'i2c_lib.py' and 'lcddriver.py' into a single library
# added bits and pieces from various sources
# By DenisFromHR (Denis Pleic)
# 2015-02-10, ver 0.1

"""

# Cobbled together from former TinkerAccess LcdApi.py and
# http://www.circuitbasics.com/raspberry-pi-i2c-lcd-set-up-and-programming/
# by kso512 (Chris Lindbergh) with lots of teaching from Erick McQueen
# 2017-05


# The wiring for the LCD with backpack is as follows:
#   LCD | Pin on Raspberry Pi
#   ====|====================
#   GND | 6 (GND)
#   VCC | 4 (5V)
#   SDA | 3 (GPIO2, SDA1 I2C)
#   SCL | 5 (GPIO3, SCL1 I2C)


import time
import logging
import I2CApi
from ClientOptionParser import ClientOption

# LCD Address
LCD_BACKPACK_ADDRESS = 0x27
LCD_SERLCD_ADDRESS = 0x72

# LCD display init sleep delay
LCD_INITDELAY = 0.2

# SerLCD command characters
SERLCD_SPECIAL_CMD = 254
SERLCD_SETTING_CMD = 0x7C

# SerLCD special commands
SERLCD_RETURNHOME = 0x02
SERLCD_ENTRYMODESET = 0x04
SERLCD_DISPLAYCONTROL = 0x08
SERLCD_CURSORSHIFT = 0x10
SERLCD_SETDDRAMADDR = 0x80

# SerLCD setting commands (partial list)
SERLCD_CONTRAST = 0x18
SERLCD_RGB_BACKLIGHT = 0x2B
SERLCD_CLEARDISPLAY = 0x2D

# LCD commands for backpack only
LCD_CLEARDISPLAY = 0x01
LCD_RETURNHOME = 0x02
LCD_ENTRYMODESET = 0x04
LCD_DISPLAYCONTROL = 0x08
LCD_CURSORSHIFT = 0x10
LCD_FUNCTIONSET = 0x20
LCD_SETCGRAMADDR = 0x40
LCD_SETDDRAMADDR = 0x80

# flags for LCD display entry mode, common to backpack and SerLCD
LCD_ENTRYRIGHT = 0x00
LCD_ENTRYLEFT = 0x02
LCD_ENTRYSHIFTINCREMENT = 0x01
LCD_ENTRYSHIFTDECREMENT = 0x00

# flags for LCD display on/off control, common to backpack and SerLCD
LCD_DISPLAYON = 0x04
LCD_DISPLAYOFF = 0x00
LCD_CURSORON = 0x02
LCD_CURSOROFF = 0x00
LCD_BLINKON = 0x01
LCD_BLINKOFF = 0x00

# flags for LCD display/cursor shift, common to backpack and SerLCD
LCD_DISPLAYMOVE = 0x08
LCD_CURSORMOVE = 0x00
LCD_MOVERIGHT = 0x04
LCD_MOVELEFT = 0x00

# flags for LCD function set, backpack only
LCD_8BITMODE = 0x10
LCD_4BITMODE = 0x00
LCD_2LINE = 0x08
LCD_1LINE = 0x00
LCD_5x10DOTS = 0x04
LCD_5x8DOTS = 0x00

# flags for LCD backlight control, backpack only
LCD_BACKLIGHT = 0x08
LCD_NOBACKLIGHT = 0x00

En = 0b00000100 # LCD Enable bit
Rw = 0b00000010 # LCD Read/Write bit
Rs = 0b00000001 # LCD Register select bit

class LcdApi(object):

    # Initialize objects and LCD
    def __init__(self, opts, init=True):
        self.__logger = logging.getLogger(__name__)
        self.__opts = opts
        self.__init = init
        self.__serlcd = self.__opts.get(ClientOption.DISPLAY_SERLCD)

        if self.__serlcd:
            # SerLCD address
            self.lcd_device = I2CApi.i2c_device(LCD_SERLCD_ADDRESS)
        else:
            # Backpack address
            self.lcd_device = I2CApi.i2c_device(LCD_BACKPACK_ADDRESS)

        if not self.__serlcd:
            # Only for backpack
            self.lcd_write(0x03)
            self.lcd_write(0x03)
            self.lcd_write(0x03)
            self.lcd_write(0x02)

        if (self.__init):
            self.__init__display()

    # Fulfill context management requirements
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    # Initialize display
    def __init__display(self):

        if self.__serlcd:
            # SerLCD
            self.serlcd_write(SERLCD_SPECIAL_CMD, [LCD_DISPLAYCONTROL | LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF])
            self.serlcd_write(SERLCD_SPECIAL_CMD, [LCD_ENTRYMODESET | LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT])
            self.serlcd_write(SERLCD_SETTING_CMD, [SERLCD_CLEARDISPLAY])
            #self.serlcd_write(SERLCD_SETTING_CMD, [SERLCD_RGB_BACKLIGHT, 0, 0, 0])

        else:
            # Backpack
            self.lcd_write(LCD_FUNCTIONSET | LCD_2LINE | LCD_5x8DOTS | LCD_4BITMODE)
            self.lcd_write(LCD_DISPLAYCONTROL | LCD_DISPLAYON)
            self.lcd_write(LCD_CLEARDISPLAY)
            self.lcd_write(LCD_ENTRYMODESET | LCD_ENTRYLEFT)
            
        time.sleep(LCD_INITDELAY)
        
    # Write a command to SerLCD
    def serlcd_write(self, cmd, data):
        self.lcd_device.write_block_data(cmd, data)
        
    # Write a command to backpack LCD
    def lcd_write(self, cmd, mode=0):
        self.lcd_write_four_bits(mode | (cmd & 0xF0))
        self.lcd_write_four_bits(mode | ((cmd << 4) & 0xF0))

    # Write four bits to the backpack LCD
    def lcd_write_four_bits(self, data):
        self.lcd_device.write_cmd(data | LCD_BACKLIGHT)
        self.lcd_strobe(data)

    # clocks EN to latch command for backpack
    def lcd_strobe(self, data):
        self.lcd_device.write_cmd(data | En | LCD_BACKLIGHT)
        time.sleep(.0005)
        self.lcd_device.write_cmd(((data & ~En) | LCD_BACKLIGHT))
        time.sleep(.0001)

    # write a character to lcd (or character rom)
    # 0x09: backlight | RS=DR< works!
    def lcd_write_char(self, charvalue, mode=1):
        self.lcd_write_four_bits(mode | (charvalue & 0xF0))
        self.lcd_write_four_bits(mode | ((charvalue << 4) & 0xF0))

    # put string function with optional char positioning
    def lcd_display_string(self, string, line=1, pos=0):
        if line == 1:
            pos_new = pos
        elif line == 2:
            pos_new = 0x40 + pos
        elif line == 3:
            pos_new = 0x14 + pos
        elif line == 4:
            pos_new = 0x54 + pos

        if self.__serlcd:
            # SerLCD
            self.serlcd_write(SERLCD_SPECIAL_CMD, [SERLCD_SETDDRAMADDR + pos_new] + [ord(char) for char in string])
            
        else:
            # Backpack
            self.lcd_write(0x80 + pos_new)

            for char in string:
                self.lcd_write(ord(char), Rs)

    # clear lcd and set to home
    def lcd_clear(self):
        self.lcd_write(LCD_CLEARDISPLAY)
        self.lcd_write(LCD_RETURNHOME)

    # define backlight on/off (lcd.backlight(1); off= lcd.backlight(0)
    def backlight(self, state): # for state, 1 = on, 0 = off
        if state == 1:
            self.lcd_device.write_cmd(LCD_BACKLIGHT)
        elif state == 0:
            self.lcd_device.write_cmd(LCD_NOBACKLIGHT)

    # add custom characters (0 - 7)
    def lcd_load_custom_chars(self, fontdata):
        self.lcd_write(0x40);
        for char in fontdata:
            for line in char:
                self.lcd_write_char(line)

    # Write two lines to the LCD, called from outside
    def write(self, first_line, second_line):
	self.lcd_display_string(first_line, 1, 0)
	self.lcd_display_string(second_line, 2, 0)

    # Set the RGB backlight (SerLCD only)
    def rgb_backlight(self, red, green, blue):
        if self.__serlcd:
            self.serlcd_write(SERLCD_SETTING_CMD, [SERLCD_RGB_BACKLIGHT, (0,255)[red], (0,255)[green], (0,255)[blue]])




import threading

from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientOption import ClientOption
from CommandExecutor import CommandExecutor


class AutoUpdateTimer(object):

    def __init__(self, client, opts):
        self.__opts = opts
        self.__client = client
        self.__auto_update_timer = None
        self.__logger = ClientLogger.setup(opts)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__cancel_auto_update_timer()

    def __cancel_auto_update_timer(self):
        if self.__auto_update_timer:
            self.__auto_update_timer.cancel()

        self.__auto_update_timer = None

    # noinspection PyBroadException
    def __auto_update_timer_tick(self):
        if not self.__client.is_in_use():
            try:
                CommandExecutor().execute_commands([
                    '{0} update'.format(PackageInfo.pip_package_name)
                ])
            except Exception:
                pass

        self.__start_auto_update_timer()

    def __start_auto_update_timer(self):
        self.__cancel_auto_update_timer()

        auto_update = self.__opts.get(ClientOption.AUTO_UPDATE)
        if auto_update:
            auto_update_interval = self.__opts.get(ClientOption.AUTO_UPDATE_INTERVAL) * 60
            self.__auto_update_timer = threading.Timer(
                auto_update_interval,
                self.__auto_update_timer_tick
            )
            self.__auto_update_timer.start()

    def start(self):
        self.__start_auto_update_timer()



import os
import tempfile
import subprocess

from ClientLogger import ClientLogger


class CommandExecutor(object):

    def __init__(self, phase=None):
        #TODO: going to need to pass in opts for clientLogger
        self.__logger = ClientLogger.setup(phase=phase)

    def ensure_execute_permission(self, path):
        try:
            os.chmod(path, 0755)
        except Exception as e:
            self.__logger.exception(e)
            raise e

    def execute_commands(self, commands):

        # I suppose an explanation is warranted here...
        # Unfortunately we cannot execute these commands directly from python due to the fact that the
        # start priority 91 must be passed to the update-rc command as an integer and python converts all arguments to
        # strings which causes and exception when the update command is invoked.
        # We work around the problem by creating a temporary script file and executing that

        fd, path = tempfile.mkstemp()
        try:

            with os.fdopen(fd, 'w') as tmp:
                tmp.writelines(['#!/usr/bin/env bash\n\n'] + ['{0}\n'.format(command) for command in commands])
            self.ensure_execute_permission(path)
            self.execute_command(path)
        finally:
            os.remove(path)

    def execute_command(self, command):
        try:
            cmd = [command] + ['-evx']  # Options: http://www.tldp.org/LDP/abs/html/options.html
            cmd_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout_data, stderr_data = cmd_process.communicate()
            if cmd_process.returncode != 0:
                for ln in stderr_data.splitlines(True):
                    self.__logger.error(ln)
                raise RuntimeError('{0} command failed.\n'.format(cmd))
            else:
                for ln in stdout_data.splitlines(True):
                    self.__logger.debug(ln)
        except RuntimeError as e:
            raise e
        except Exception as e:
            self.__logger.exception(e)
            raise e



import logging
from Command import Command
from PackageInfo import PackageInfo


class CommandHandler(object):

    def __init__(self, opts=None, args=None):
        self.__logger = logging.getLogger(__name__)
        self.__args = args or []
        self.__opts = opts or {}
        self.__handlers = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def handle_command(self):

        cmd = Command(self.__args[0].lower() if len(self.__args) >= 1 and len(self.__args[0].lower()) >= 1 else None)
        for (command, call_back) in self.__handlers:
            if command is cmd:
                try:
                    return call_back(opts=self.__opts, args=self.__args)
                except Exception as e:
                    self.__logger.debug('%s failed to handle the command.', PackageInfo.pip_package_name)
                    self.__logger.exception(e)
                    raise e

    def on(self, command, call_back):
        self.__handlers.append((command, call_back,))
